/*
 * arch/mips/kernel/traps.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

/*
 * 'traps.c' handles hardware traps and faults after we have saved some
 * state in 'asm.s'. Currently mostly a debugging-aid, will be extended
 * to mainly kill the offending process (probably by giving it a signal,
 * but possibly by killing it outright if necessary).
 *
 * FIXME: This is the place for a fpu emulator.
 *
 * Modified for R3000 by Paul M. Antoine, 1995, 1996
 */
#include <linux/mm.h>

#include <asm/jazz.h>
#include <asm/vector.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/bootinfo.h>
#include <asm/watch.h>

#undef CONF_DEBUG_EXCEPTIONS

static inline void console_verbose(void)
{
	extern int console_loglevel;
	console_loglevel = 15;
}

/*
 * Machine specific interrupt handlers
 */
extern asmlinkage void acer_pica_61_handle_int(void);
extern asmlinkage void decstation_handle_int(void);
extern asmlinkage void deskstation_rpc44_handle_int(void);
extern asmlinkage void deskstation_tyne_handle_int(void);
extern asmlinkage void mips_magnum_4000_handle_int(void);

extern asmlinkage void handle_mod(void);
extern asmlinkage void handle_tlbl(void);
extern asmlinkage void handle_tlbs(void);
extern asmlinkage void handle_adel(void);
extern asmlinkage void handle_ades(void);
extern asmlinkage void handle_ibe(void);
extern asmlinkage void handle_dbe(void);
extern asmlinkage void handle_sys(void);
extern asmlinkage void handle_bp(void);
extern asmlinkage void handle_ri(void);
extern asmlinkage void handle_cpu(void);
extern asmlinkage void handle_ov(void);
extern asmlinkage void handle_tr(void);
extern asmlinkage void handle_vcei(void);
extern asmlinkage void handle_fpe(void);
extern asmlinkage void handle_vced(void);
extern asmlinkage void handle_watch(void);
extern asmlinkage void handle_reserved(void);

static char *cpu_names[] = CPU_NAMES;

unsigned long page_colour_mask;
unsigned int watch_available = 0;

static void (*ibe_board_handler)(struct pt_regs *regs);
static void (*dbe_board_handler)(struct pt_regs *regs);

int kstack_depth_to_print = 24;

/*
 * These constant is for searching for possible module text segments.
 * MODULE_RANGE is a guess of how much space is likely to be vmalloced.
 */
#define MODULE_RANGE (8*1024*1024)

void die_if_kernel(char * str, struct pt_regs * regs, long err)
{
	int	i;
	int	*stack;
	u32	*sp, *pc, addr, module_start, module_end;
	extern	char start_kernel, _etext;

#ifdef __R4000__
	/*
	 * Just return if in user mode.
	 */
	if (!(regs->cp0_status & 0x18))
		return;
#else
	if (!((regs)->cp0_status & 0x4))
		return;
#endif

	/*
	 * Yes, these double casts are required ...
	 */
	sp = (u32 *)(unsigned long)regs->regs[29];
	pc = (u32 *)(unsigned long)regs->cp0_epc;

	console_verbose();
	printk("%s: %08lx\n", str, err );

	show_regs(regs);

	/*
	 * Dump the stack
	 */
	if (STACK_MAGIC != *(u32 *)current->kernel_stack_page)
		printk("Corrupted stack page\n");
	printk("Process %s (pid: %d, stackpage=%08lx)\nStack: ",
		current->comm, current->pid, current->kernel_stack_page);
	for(i=0;i<5;i++)
		printk("%08x ", *sp++);
	stack = (int *) sp;
	for(i=0; i < kstack_depth_to_print; i++) {
		if (((u32) stack & (PAGE_SIZE -1)) == 0)
			break;
		if (i && ((i % 8) == 0))
			printk("\n       ");
		printk("%08x ", get_user(stack++));
	}
	printk("\nCall Trace: ");
	stack = (int *)sp;
	i = 1;
	module_start = VMALLOC_START;
	module_end = module_start + MODULE_RANGE;
	while (((u32)stack & (PAGE_SIZE -1)) != 0) {
		addr = get_user(stack++);
		/*
		 * If the address is either in the text segment of the
		 * kernel, or in the region which contains vmalloc'ed
		 * memory, it *may* be the address of a calling
		 * routine; if so, print it so that someone tracing
		 * down the cause of the crash will be able to figure
		 * out the call path that was taken.
		 */
		if (((addr >= (u32) &start_kernel) &&
		     (addr <= (u32) &_etext)) ||
		    ((addr >= module_start) && (addr <= module_end))) {
			if (i && ((i % 8) == 0))
				printk("\n       ");
			printk("%08x ", addr);
			i++;
		}
	}

	printk("\nCode : ");
	if ((!verify_area(VERIFY_READ, pc, 5 * sizeof(*pc)) ||
	     KSEGX(pc) == KSEG0 ||
	     KSEGX(pc) == KSEG1) &&
	    (((unsigned long) pc & 3) == 0))
	{
		for(i=0;i<5;i++)
			printk("%08x ", *pc++);
		printk("\n");
	}
	else
		printk("(Bad address in epc)\n");
	while(1);
	do_exit(SIGSEGV);
}

/*
 * Compute the return address and do emulate branch and instruction
 * simulation, if required.
 */
inline int compute_return_epc(struct pt_regs *regs, char *exc)
{
	/* All other MIPS operating systems have the capability to execute
	   instructions like break and Tcc even in branch delay slots.  Neither
	   GCC nor GAS or GNU ld produce such code under normal circumstances.
	   So we're completly safe without this capability.  Later on we'll
	   implement it for support of ptrace(2) and compatibility. */
	if (regs->cp0_cause & CAUSEF_BD)
	{
		printk("%s: %s exception in branch delay slot - killed.\n",
		       current->comm, exc);
		/*
		 * This exception is in a branch delay slot.  Since we don't
		 * yet do branch simulation we might get a looping
		 * exception.  So better zap the process unconditional.
		 */
		do_exit(SIGILL);
		return -EFAULT;
	}
	if (regs->cp0_epc & 3)
	{
		/*
		 * User program has an unaligned epc.
		 * This does not happen, so this check is pure paranoia.
		 */
		printk("%s: unaligned epc in compute_return_epc() - sending SIGSEGV.\n",
		       current->comm);
		send_sig(SIGSEGV, current, 1);
		return -EFAULT;
	}
	regs->cp0_epc += 4;

	return 0;
}

/*
 * The ibe/dbe exceptions are signaled by onboard hardware and should get
 * a board specific handlers to get maximum available information. Bus
 * errors are always symptom of hardware malfunction or a kernel error.
 * We should try to handle this case a bit more gracefully than just
 * zapping the process ...
 */
static void jazz_be_board_handler(struct pt_regs *regs)
{
	u32	jazz_is, jazz_ia;

	/*
	 * Give some debugging aid ...
	 */
	jazz_is = r4030_read_reg32(JAZZ_R4030_IRQ_SOURCE);
	jazz_ia = r4030_read_reg32(JAZZ_R4030_INVAL_ADDR);
	printk("Interrupt Source         == %08x\n", jazz_is);
	printk("Invalid Address Register == %08x\n", jazz_ia);
	show_regs(regs);

	/*
	 * Assume it would be too dangerous to continue ...
	 */
	send_sig(SIGBUS, current, 1);
}

static void default_be_board_handler(struct pt_regs *regs)
{
	/*
	 * Assume it would be too dangerous to continue ...
	 */
	send_sig(SIGBUS, current, 1);
}

void do_ibe(struct pt_regs *regs)
{
	ibe_board_handler(regs);
}

void do_dbe(struct pt_regs *regs)
{
	dbe_board_handler(regs);
}

void do_ov(struct pt_regs *regs)
{
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	send_sig(SIGFPE, current, 1);
}

void do_fpe(struct pt_regs *regs, unsigned int fcr31)
{
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	send_sig(SIGFPE, current, 1);
}

static inline int get_insn_opcode(struct pt_regs *regs, unsigned int *opcode)
{
	unsigned int *epc;

	epc = (unsigned int *) (unsigned long) regs->cp0_epc;
	if (regs->cp0_cause & CAUSEF_BD)
		epc += 4;

	if (verify_area(VERIFY_READ, epc, 4))
	{
		send_sig(SIGSEGV, current, 1);
		return 1;
	}
	*opcode = *epc;

	return 0;
}

void do_bp_and_tr(struct pt_regs *regs, char *exc, unsigned int trapcode)
{
	/*
	 * (A short test says that IRIX 5.3 sends SIGTRAP for all break
	 * insns, even for break codes that indicate arithmetic failures.
	 * Wiered ...)
	 */
	send_sig(SIGTRAP, current, 1);
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	if (compute_return_epc(regs, exc))
		return;
}

void do_bp(struct pt_regs *regs)
{
	unsigned int opcode, bcode;

	/*
	 * There is the ancient bug in the MIPS assemblers that the break
	 * code starts left to bit 16 instead to bit 6 in the opcode.
	 * Gas is bug-compatible ...
	 */
	if (get_insn_opcode(regs, &opcode))
		return;
	bcode = ((opcode >> 16) & ((1 << 20) - 1));

	do_bp_and_tr(regs, "bp", bcode);
}

void do_tr(struct pt_regs *regs)
{
	unsigned int opcode, bcode;

	if (get_insn_opcode(regs, &opcode))
		return;
	bcode = ((opcode >> 6) & ((1 << 20) - 1));

	do_bp_and_tr(regs, "tr", bcode);
}

void do_ri(struct pt_regs *regs)
{
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	if (compute_return_epc(regs, "ri"))
		return;
	send_sig(SIGILL, current, 1);
}

void do_cpu(struct pt_regs *regs)
{
	unsigned int cpid;

	cpid = (regs->cp0_cause >> CAUSEB_CE) & 3;
	if (cpid == 1)
	{
		regs->cp0_status |= ST0_CU1;
		return;
	}

	send_sig(SIGILL, current, 1);
}

void do_vcei(struct pt_regs *regs)
{
	/*
	 * Only possible on R4[04]00[SM]C. No handler because I don't have
	 * such a cpu.  Theory says this exception doesn't happen.
	 */
	panic("Caught VCEI exception - should not happen");
}

void do_vced(struct pt_regs *regs)
{
	/*
	 * Only possible on R4[04]00[SM]C. No handler because I don't have
	 * such a cpu.  Theory says this exception doesn't happen.
	 */
	panic("Caught VCE exception - should not happen");
}

void do_watch(struct pt_regs *regs)
{
	/*
	 * We use the watch exception where available to detect stack
	 * overflows.
	 */
	show_regs(regs);
	panic("Caught WATCH exception - probably caused by stack overflow.");
}

void do_reserved(struct pt_regs *regs)
{
	/*
	 * Game over - no way to handle this if it ever occurs.
	 * Most probably caused by a new unknown cpu type or
	 * after another deadly hard/software error.
	 */
	panic("Caught reserved exception - should not happen.");
}

static void watch_init(unsigned long cputype)
{
	switch(cputype) {
	case CPU_R10000:
	case CPU_R4000MC:
	case CPU_R4400MC:
	case CPU_R4000SC:
	case CPU_R4400SC:
	case CPU_R4000PC:
	case CPU_R4400PC:
	case CPU_R4200:
     /* case CPU_R4300: */
		set_except_vector(23, handle_watch);
		watch_available = 1;
		break;
	}
}

void trap_init(void)
{
	unsigned long	i;

	if (mips_machtype == MACH_MIPS_MAGNUM_4000) 
		EISA_bus = 1;

	/*
	 * Setup default vectors
	 */
	for (i=0;i<=31;i++)
		set_except_vector(i, handle_reserved);

	watch_init(mips_cputype);
	/*
	 * Handling the following exceptions depends mostly of the cpu type
	 */
	switch(mips_cputype) {
	case CPU_R10000:
		/*
		 * The R10000 is in most aspects similar to the R4400.  It
		 * should get some special optimizations.
		 */
		write_32bit_cp0_register(CP0_FRAMEMASK, 0);
		set_cp0_status(ST0_XX, ST0_XX);
		/*
		 * Actually this mask stands for only 16k cache.  This is
		 * correct since the R10000 has multiple ways in it's cache.
		 */
		page_colour_mask = 0x3000;
		/*
		 * The R10k might even work for Linux/MIPS - but we're paranoid
		 * and refuse to run until this is tested on real silicon
		 */
		panic("CPU too expensive - making holiday in the ANDES!");
		break;
	case CPU_R4000MC:
	case CPU_R4400MC:
	case CPU_R4000SC:
	case CPU_R4400SC:
		/*
		 * Handlers not implemented yet.  Shouldn't ever be used -
		 * otherwise it's a bug in the Linux/MIPS kernel, anyway.
		 */
		set_except_vector(14, handle_vcei);
		set_except_vector(31, handle_vced);
	case CPU_R4000PC:
	case CPU_R4400PC:
	case CPU_R4200:
     /* case CPU_R4300: */
     /* case CPU_R4640: */
	case CPU_R4600:
		set_except_vector(1, handle_mod);
		set_except_vector(2, handle_tlbl);
		set_except_vector(3, handle_tlbs);
		set_except_vector(4, handle_adel);
		set_except_vector(5, handle_ades);
		/*
		 * The following two are signaled by onboard hardware and
		 * should get board specific handlers to get maximum
		 * available information.
		 */
		set_except_vector(6, handle_ibe);
		set_except_vector(7, handle_dbe);

		set_except_vector(8, handle_sys);
		set_except_vector(9, handle_bp);
		set_except_vector(10, handle_ri);
		set_except_vector(11, handle_cpu);
		set_except_vector(12, handle_ov);
		set_except_vector(13, handle_tr);
		set_except_vector(15, handle_fpe);

		/*
		 * Compute mask for page_colour().  This is based on the
		 * size of the data cache.
		 */
		i = read_32bit_cp0_register(CP0_CONFIG);
		i = (i >> 26) & 7;
		page_colour_mask = 1 << (12 + i);
		break;
	case CPU_R6000:
	case CPU_R6000A:
#if 0
		/* The R6000 is the only R-series CPU that features a machine
		 * check exception (similar to the R4000 cache error) and
		 * unaligned ldc1/sdc1 exception.  The handlers have not been
		 * written yet.  Well, anyway there is no R6000 machine on the
		 * current list of targets for Linux/MIPS.
		 */
		set_except_vector(14, handle_mc);
		set_except_vector(15, handle_ndc);
#endif
	case CPU_R2000:
	case CPU_R3000:
	case CPU_R3000A:
		/*
		 * Actually don't know about these, but let's guess - PMA
		 */
		set_except_vector(1, handle_mod);
		set_except_vector(2, handle_tlbl);
		set_except_vector(3, handle_tlbs);
		set_except_vector(4, handle_adel);
		set_except_vector(5, handle_ades);
		/*
		 * The Data Bus Error/ Instruction Bus Errors are signaled
		 * by external hardware.  Therefore these two expection have
		 * board specific handlers.
		 */
		set_except_vector(6, handle_ibe);
		set_except_vector(7, handle_dbe);
		switch(mips_machtype) {
			case MACH_ACER_PICA_61:
			case MACH_MIPS_MAGNUM_4000:
				ibe_board_handler = jazz_be_board_handler;
				dbe_board_handler = jazz_be_board_handler;
				break;
			default:
				ibe_board_handler = default_be_board_handler;
				dbe_board_handler = default_be_board_handler;
				break;
		}

		set_except_vector(8, handle_sys);
		set_except_vector(9, handle_bp);
		set_except_vector(10, handle_ri);
		set_except_vector(11, handle_cpu);
		set_except_vector(12, handle_ov);
		set_except_vector(13, handle_tr);
		set_except_vector(15, handle_fpe);

		/*
		 * Compute mask for page_colour().  This is based on the
		 * size of the data cache.  Does the size of the icache
		 * need to be accounted for?
		 *
		 * FIXME: is any of this necessary for the R3000, which
		 *        doesn't have a config register?
		 *        (No, the R2000, R3000 family has a physical indexed
		 *        cache and doesn't need this braindamage.)
		i = read_32bit_cp0_register(CP0_CONFIG);
		i = (i >> 26) & 7;
		page_colour_mask = 1 << (12 + i);
		 */
		break;
	case CPU_R3041:
	case CPU_R3051:
	case CPU_R3052:
	case CPU_R3081:
	case CPU_R3081E:
	case CPU_R8000:
        case CPU_R5000:
		printk("Detected unsupported CPU type %s.\n",
			cpu_names[mips_cputype]);
		panic("Can't handle CPU");
		break;

	case CPU_UNKNOWN:
	default:
		panic("Unknown CPU type");
	}

	/*
	 * The interrupt handler mostly depends of the board type.
	 */
	set_except_vector(0, feature->handle_int);
}
